"""
Module `chatette.units.slot`
Cotains the definition and rule contents characterizing slots (i.e. entities).
"""

from .definition import (SlotDefinition, )
from .rule_content import (DummySlotValRuleContent, SlotRuleContent)
