"""
Module `chatette.units.intent`
Contains the definition and the content rules charcterizing intents.
"""

from .definition import (IntentDefinition, )
from .example import (IntentExample, )
from .rule_content import (IntentRuleContent, )
