"""
Module `chatette.units.word`
Contains the rule contents characterizing words and word groups.
"""

from .group_rule_content import (GroupWordRuleContent, )
from .rule_content import (WordRuleContent, )
